import os
from functools import reduce
from skimage import io
import numpy as np
import pandas as pd
import glob
from skimage.filters import laplace
from skimage.util import dtype_limits
from .threshold import otsuThreshold
from typing import List, Dict

import PIL # this is needed to read in the giant images with skimage.io.read_collection
PIL.Image.MAX_IMAGE_PIXELS = 933120000

def getImageStats(image: np.ndarray, out_print: bool = True, save: bool = False, result_prefix: str = "results", add_props: Dict = {}) -> pd.DataFrame:
    """Returns general stats about the input image.

    Parameters
    ----------
    image : np.ndarray
        input image
    out_print : bool
        Determines whether to print out the results to stdout, default = False
    save : bool
        Determines whether to save the results to a csv, default = False
    result_prefix : str
        prefix to use when saving the csv: output file name will be {result_prefix}.csv

    Returns
    -------
    pd.DataFrame

    """
    attribute_col = []
    value_col = []

    attribute_col.append("Shape")
    value_col.append(image.shape)

    attribute_col.append("Dtype")
    value_col.append(image.dtype)

    attribute_col.append("Dtype range")
    value_col.append(dtype_limits(image))

    attribute_col.append("Actual range")
    value_col.append((np.amin(image), np.amax(image)))

    attribute_col.append("Mean")
    value_col.append(np.mean(image))

    attribute_col.append("Median")
    value_col.append(np.median(image))

    attribute_col.append("Laplace std")
    value_col.append(np.std(laplace(image)))

    otsu_threshold = otsuThreshold(image)
    attribute_col.append("Otsu threshold")
    value_col.append(otsu_threshold)

    above_otsu = np.sum(np.where(image > otsu_threshold, 1, 0))
    below_otsu =(image.shape[0] * image.shape[1]) - above_otsu
    perc_above = above_otsu / (image.shape[0] * image.shape[1]) * 100

    attribute_col.append("Pixels above otsu")
    value_col.append(above_otsu)

    attribute_col.append("Pixels below otsu")
    value_col.append(below_otsu)

    attribute_col.append("Percentage above otsu")
    value_col.append(perc_above)

    if add_props:
        for key, value in add_props.items():
            attribute_col.append(key)
            value_col.append(value)

    # Make result dataframe
    zipped = list(zip(attribute_col, value_col))
    result_df = pd.DataFrame(zipped, columns=['Attribute', f'{result_prefix}'], index=None)
    result_df.index=result_df['Attribute']
    result_df = result_df.drop('Attribute', axis=1)

    if out_print:
        print(result_df.to_markdown())
    if save:
        result_df.T.to_csv(f"{result_prefix}.csv", index=None)
    return result_df.T

def compareImageStats(glob_pattern: str = None, result_prefix = "result", add_props: Dict = {}) -> pd.DataFrame:
    """Create a dataframe comparing the general stats of a collection of images, defined by a glob pattern.

    Parameters
    ----------
    glob_pattern : str
        glob_pattern
    result_prefix :
        prefix to use when saving the csv: output file name will be {result_prefix}.html

    Returns
    -------
    pd.DataFrame

    """
    
    if isinstance(glob_pattern, list):
        image_path_list = glob_pattern
    else:
        image_path_list = glob.glob(glob_pattern)
    
    name_image_dict = {os.path.basename(file): io.imread(file) for file in image_path_list}
    # image_list = io.imread_collection(glob_pattern)

    df_list  = []
    for k,v in name_image_dict.items():
        df_list.append((getImageStats(v, result_prefix = k, save=False, out_print=False, add_props = add_props))) # Transpose to make concatenating possible

    # merged_df = reduce(lambda x, y: pd.merge(x, y, on = 'Attribute'), df_list)
    merged_df = pd.concat(df_list)
    merged_df = merged_df.sort_index() # sort for readability
    merged_df.to_csv(f"{result_prefix}.csv")
    return merged_df
        



from skimage import io
import pandas as pd
from skimage.draw import disk, rectangle
from skimage.transform import rotate, resize, warp
from pyxelperfect.decorators import  measureTime
import SimpleITK as sitk
import numpy as np
import math
import matplotlib.pyplot as plt
import os
import ast

"""
This module assumes 'point': len(tuple) == 2 with row and col being the fir and second element respectively.
All functions are structured in a way that they take only 2 arguments, point and value, where value is everything the function needs to do is thing.
This is to make sure that registerPoint can handle an arbitrary amount of different instructions, at the expense of readability.
"""

def loadRegisterKwargsFromCsv(path_to_csv):
    ## Ordered processing steps will be a list of single element dicts, where key = step and value = value. registerPoint will have to be adapted with a mapping that maps step to their respective functions
    def readFlip(value):
        return ast.literal_eval(value)
    def readBbox(value):
        return ast.literal_eval(value)
    # usually when transforming a point it's for transforming a point belonging to the target coordinate axis, so we assume we want an inverse transform
    def readTransform(value):
        if "mha" in value:
            return sitk.DisplacementFieldTransform(sitk.ReadImage(value))
        else:
            return sitk.ReadTransform(value).GetInverse()
    def readRotation(value):
        rotation = ast.literal_eval(value)
        center = rotation[0]
        angle = rotation[1]
        if len(rotation) > 2:
            translation = rotation[2]
        else:
            translation = (0,0)
        return {"center": center, "angle": angle, "translation": translation}
    def readZoomFactors(value):
        return ast.literal_eval(value)

    def readTranslation(value):
        return ast.literal_eval(value)

    functions = {"flip": readFlip, "bbox": readBbox, "rotation":readRotation, "zoom_factors": readZoomFactors, "transform": readTransform, "translation": readFlip}


    df = pd.read_csv(path_to_csv)
    ordered_processing_steps = []
    for row in df.itertuples():
        # The value of the single element dict is already the parsed argument, as to not load in transforms multiple times when registering multiple points
        ordered_processing_steps.append({row.step: functions[row.step](row.value)})

    return ordered_processing_steps


def registerPoint(point, args_list):
    args_dict = {"flip": flipPoint, "rotation": rotatePointDict, "bbox": transformPointsWithBbox, "zoom_factors": resizePointWithTransform, "transform": sitkTransformPoint, "translation": translatePoint}
    for arg in args_list:
        for key, value in arg.items():
            # arg = dict and will only have one item
            point = args_dict[key](point, value)
            if isinstance(point, type(None)):
                return (0,0)
    return point

def findCenter(x_shape):
    middle = x_shape[0] / 2, x_shape[1] / 2
    return middle

def rotatePointDict(point, arg_dict: dict):
    # Duplicate of rotatePoint but with takes input as a dict. This just to facilitate the function dict used in registerPoint

    origin, angle= arg_dict["center"], arg_dict["angle"]
    try:
        translation = arg_dict["translation"]
    except:
        translation = (0,0)

    point = (point[0] - translation[0], point[1] - translation[1])

    if isinstance(angle, int):
        angle = math.radians(angle)
    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return (qx, qy)

def rotatePoint(point, origin, angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.
    If angle is an integer, it is assumed to be degrees and will be converted to radians
    """

    if isinstance(angle, int):
        angle = math.radians(angle)

    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return (qx, qy)

def cutBboxFromImage(bbox, image: np.ndarray):
    min_row, min_col, max_row, max_col = bbox

    cut_image = image[min_row:max_row, min_col:max_col]
    return cut_image

def getShapeFromBbox(bbox):
    min_row, min_col, max_row, max_col = bbox
    return (max_row, max_col)

def point2int(point):
    return tuple([round(el) for el in point])

def transformPointsWithBbox(point, bbox, force = False):
    min_row, min_col, max_row, max_col = bbox
    if min_row <= point[0] < max_row and min_col <= point[1] < max_col:
        new_row = point[0] - min_row
        new_col = point[1] - min_col
        return (new_row, new_col)
    else:
        if not force:
            return None
        else:
            # If it doesn't fit, we find the closest point to the bbox that does fit
            closest_row = min(max(min_row, point[0]), max_row)
            closest_col = min(max(min_col, point[1]), max_col)
        
            return closest_row - min_row, closest_col - min_col

def calculateResizeZoomFactors(original_shape, target_shape):
    """
    Calculates the zoom factors necessary to go from original_shape to target shape
    """
    tform = [] 
    for i in range(len(original_shape)):
        tform.append(target_shape[i] / original_shape[i])
    zoom_factors = np.array(tform)
    return zoom_factors

def resizePointWithTransform(point, zoom_factors):
    """resizePointWithTransform.
    Parameters
    ----------
    point :
        2D point in form of a tuple
    zoom_factors :
        transform in the form of the zoom factors used in the ndi.zoom function by scipy
    """

    point = np.array([point]) # add a second artificial dimension to it
    new_point = np.array(zoom_factors) * point
    new_point = (float(new_point[0,0]), float(new_point[0,1]))
    return new_point

def reflectPoint180degrees(point, origin):
    new_point = 2*origin[0] - point[0], 2*origin[1] - point[1]
    return new_point

def flipPoint(point, shape):
    """
    Flips a point along a specific axis. 
    shape = [n_rows, n_cols, axis],
    where if axis  == 0: no flip
    axis == 1: flip along the rows | vertically | mirror over a horizontal line 
    axis == 2: flip along the cols | horizonatlly | mirror over a vertical line 
    axis == 3: flip along both rows and cols | vertically  and horizontally | mirror over both a horizontal and vertical line 
    """
    img_shape, axis = shape[:2], shape[2]
    # img_shape = shape[:2]
    
    if axis == 0: 
        return point
    elif axis == 1:
        new_row = (img_shape[0] - point[0]) - 1
        new_col = point[1]
        new_point = (new_row, new_col)

    elif axis == 2:
        new_col = (img_shape[1] - point[1]) -1
        new_row = point[0]
        new_point = (new_row, new_col)

    elif axis == 3:
        new_row = (img_shape[0] - point[0]) - 1
        new_col = (img_shape[1] - point[1]) -1
        new_point = (new_row, new_col)
    return new_point

def sitkTransformPoint(point, sitk_transform):
    """sitkTransformPoint.
    -   sitk.transform.TransformPoint assumes (col, row) under the hood, so we invert them beforehand and invert them back after transform
    Parameters
    ----------
    point :
        tuple: (row, col)
    sitk_transform :
        sitk_transform
    """
    if isinstance(sitk_transform, str):
        sitk_transform = sitk.ReadTransform(sitk_transform)
    inverted_point = (point[1], point[0])
    transformed_point = sitk_transform.TransformPoint(inverted_point)
    return_point = (transformed_point[1], transformed_point[0])
    return return_point

def translatePoint(point, offset):
    new_point =(point[0] + offset[0] , point[1] + offset[1])
    return new_point



import numpy as np
from skimage.filters import sobel
from skimage import io
from typing import List


def maxIPstack(img_list: List[np.array]) -> np.array:
    """Takes maximum intensity projection of a list of images

    Parameters
    ----------
    img_list : List[np.array]
        list of images

    Returns
    -------
    np.array

    """
    parsed_list = img_list
    parsed_list = [img if isinstance(img, np.ndarray) else io.imread(img) for img in img_list]

    maxIP = np.maximum.reduce(parsed_list)
    return maxIP

def edf(img_list) -> np.array:
    ## Credist to the documentation of the mahotas package https://mahotas.readthedocs.io/en/latest/edf.html
    ## Important: image array has to have the z-stacks as first dimension
    if isinstance( img_list, list):
        img_array = np.zeros((len(img_list),*img_list[0].shape))
        for i, img in enumerate(img_list):
            img_array[i, :, :] = img
    else:
        img_array = img_list
    stack,h,w = img_array.shape

    focus = np.array([sobel(img) for img in img_array])
                      
    best = np.argmax(focus, 0)

    img_array = img_array.reshape((stack,-1)) # image is now (stack, nr_pixels)
    img_array = img_array.transpose() # image is now (nr_pixels, stack)
    r = img_array[np.arange(len(img_array)), best.ravel()] # Select the right pixel at each location
    r = r.reshape((h,w)) # reshape to get final result
    return r




    



import cv2
import numpy as np
from skimage.transform import resize
from skimage.morphology import white_tophat, disk
from skimage.util import img_as_uint
from skimage import io


# testing imports
import matplotlib.pyplot as plt
from skimage.draw import ellipse


def automaticBrightnessAndContrast(image: np.array, clip_hist_percent: int = 1):
    if np.amin(image) == 0 and np.amax(image) == 0:
        return image

    # Calculate grayscale histogram
    hist, _ = np.histogram(image.flatten(), 256, [0, 256])

    # Calculate cumulative distribution from the histogram
    accumulator = hist.cumsum()

    # Locate points to clip
    maximum = accumulator[-1]
    clip_hist_percent *= (maximum / 100.0)
    clip_hist_percent /= 2.0

    try:
        # Locate left cut
        minimum_gray = 0
        while accumulator[minimum_gray] < clip_hist_percent:
            minimum_gray += 1

        # Locate right cut
        maximum_gray = len(accumulator) - 1
        while accumulator[maximum_gray] >= (maximum - clip_hist_percent):
            maximum_gray -= 1
    except:
        return image

    # If the image was empty, the alpha calculation will raise an error, so just return the original image
    if maximum_gray - minimum_gray == 0:
        return image

    # Calculate alpha and beta values
    alpha = 255 / (maximum_gray - minimum_gray)
    beta = -minimum_gray * alpha

    auto_result = exposure.rescale_intensity(image, in_range=(minimum_gray, maximum_gray))
    return auto_result

def equalizeImageSize(ref_image: np.array, target_image: np.array, save = False, out_name = "") -> np.array:
    # Gross ifelse statement to check if one of the image is globally larger or smaller than the other
    ref_shape, target_shape = ref_image.shape, target_image.shape
    # if (ref_shape[0] > target_shape[0] and ref_shape[1] < target_shape[1]) or (target_shape[0] > ref_shape[0] and target_shape[1] < ref_shape[1]):
    #     return ref_image, target_image
    # else:
    target_resized = resize(target_image,ref_image.shape)
    target_resized = img_as_uint(target_resized)
    if save and out_name:
        io.imsave(f"{out_name}_resized.tif", target_resized)
    return ref_image, target_resized

def whiteTophat(image, radius):
    """Wrapper for skimage's white tophat filter

    Parameters
    ----------
    image : np.ndarray
        Image to be filtered.
    radius: int
        Radius of the morphological disk.
    """
    selem = disk(radius)
    return white_tophat(image,selem)

def changeRGBcolor(image, original_color, target_color):
    """ Change a specific color in an image to another color

    Parameters
    ----------
    image :
        input image
    original_color :
        tuple representing the rgb color to change
    target_color :
        tuple representing the rgb color to change to 
    """
    image = image.copy()
    if len(image.shape) > 3:
        raise ValueError("not an RGB image")
    if image.shape[2] > 3:
        image = image[:,:,0:3]
    # print(image[:,:,0])
    r1, g1, b1 = original_color
    r2, g2, b2 = target_color

    red, green, blue = image[:,:,0], image[:,:,1], image[:,:,2]
    mask = (red == r1) & (green == g1) & (blue == b1)
    image[:,:,:3][mask] = [r2, g2, b2]
    return image

    # io.imsave("./image_fixed.png", data, check_contrast=False)

def changeAllColors(image, target_color):
    return_image = image.copy()
    for row in range(image.shape[0]):
        for col in range(image.shape[1]):
            if image[row,col, 0] > target_color[0] and image[row,col, 1] > target_color[1] and image[row,col, 2] > target_color[2]:
                return_image[row,col] = target_color
    return return_image

def normImage(img):
    mn = np.amin(img)
    mx = np.amax(img)
    return (img - mn) * (1.0 / (mx - mn))


if __name__ == '__main__':
    from glob import glob

    for img in glob("/home/david/Documents/phd/poster/figures/figures poster/Slide*.jpg"):
    # for img in glob("/home/david/Documents/phd/poster/graphics/fwo.jpeg"):
        original_image = io.imread(img)
        # changed = changeRGBcolor(, (255,255,255), (242, 242, 242))
        # changed = changeRGBcolor(changed, (254,254,254), (242, 242, 242))
        changed = changeAllColors(original_image, (242, 242, 242))
        io.imsave(img, changed, check_contrast=False)



import cv2
import numpy as np
import pandas as pd
from skimage import io
from skimage import color
from skimage.draw import disk
import matplotlib.pyplot as plt
from skimage.color import label2rgb
import matplotlib.patches as mpatches
from matplotlib import colors
from skimage.util import img_as_ubyte
from skimage.segmentation import find_boundaries, mark_boundaries
from .manipulate import automaticBrightnessAndContrast

def makeHist(data, mn, mx, interval):
    n, bins, patches = plt.hist(x=data, bins=np.arange(mn, mx, interval)-interval/2, color='navy',
                            alpha=0.7, rwidth=0.85)

    plt.grid(axis='y', alpha=0.75)
    plt.xlabel('Nr mitf spots')
    plt.xticks(np.arange(mn,mx,interval))
    plt.ylabel('Frequency')
    maxfreq = n.max()
    plt.show()

def showSegmentation(labeled_image: np.array, original_image, extra_labeled_image=None, extra_original_image=None, out_file=""):
    colored_image_on_DAPI = color.label2rgb(labeled_image, original_image, bg_label=0)

    if extra_original_image is not None:
        viz_img = np.zeros((*original_image.shape, 3))
        viz_img[:,:,0] = original_image /  np.amax(original_image)
        viz_img[:,:,1] = extra_original_image /  np.amax(extra_original_image)
    else:
        viz_img = original_image

    fig, axs = plt.subplots(2,2, sharex=True, sharey=True)

    axs = axs.flatten()
    for ax in axs:
        ax.axis('off')

    axs[0].imshow(viz_img)
    axs[0].set_title("original image")
    axs[1].imshow(colored_image_on_DAPI)
    axs[1].set_title("Segmentation labeled image")
    axs[2].imshow(mark_boundaries(viz_img, labeled_image))
    axs[2].set_title("overlay dapi")
    if extra_labeled_image is not None:
        axs[3].imshow(mark_boundaries(viz_img, extra_labeled_image))
        axs[3].set_title("overlay cell")

    plt.show()




def showImage(image: np.array):
    plt.imshow(image)
    plt.axis("off")
    plt.show()

def createOverlay(labeled_image, image):
    max_val = np.amax(image) 
    tmp_image = image.copy()
    tmp_image[find_boundaries(labeled_image)] = max_val
    return tmp_image

def plotBoundaries(labeled_image,image):
    plt.imshow(createOverlay(labeled_image, image))
    plt.axis("off")
    plt.show()


def createComposite(img1_path, img2_path):
    A = cv2.cv2.imread(img1_path, cv2.cv2.IMREAD_ANYDEPTH)
    B = cv2.cv2.imread(img2_path, cv2.cv2.IMREAD_ANYDEPTH)
    bitDepth = A.dtype
    # print(A.dtype, B.dtype)
    zeros = np.zeros(A.shape[:2], dtype=bitDepth)

    ##Don't forget, cv2.cv2 works with a different order sometimes, so it's not RGB it's BGR
    merged = cv2.cv2.merge((A,B,zeros))
    merged = img_as_ubyte(merged)
    return merged

def evaluateRegistration(ref_image, target_image, registered_target, identifier: str = "", plot=True):
    original_situation = createComposite(ref_image, target_image)
    original_situation = img_as_ubyte(original_situation)
    registered_situation = createComposite(ref_image, registered_target)
    registered_situation = img_as_ubyte(registered_situation)

    if identifier:
        io.imsave(f"overlay_registered_{identifier}.tif", registered_situation)
    if plot:
        fig, axs = plt.subplots(1,2)
        axs[0].imshow(original_situation)
        axs[0].set_title("Before Registration")
        red_patch = mpatches.Patch(color='red', label='MS ref')
        green_patch = mpatches.Patch(color='green', label='ST target')
        axs[0].legend(handles=[red_patch, green_patch])
        axs[1].imshow(registered_situation)
        axs[1].set_title("After Registration")
        green_patch_2 = mpatches.Patch(color='green', label='ST moved')
        axs[1].legend(handles=[red_patch, green_patch_2])
        plt.show()

def plotSpatial(image, dataframe, rowname = "row", colname="col", dotsize=5, color="red", save=False, plot=True, ax=None, colormap="tab20"):
    def get_cmap(n, name='tab20'):
        ##Returns a function that maps each index in 0, 1, ..., n-1 to a distinct 
        ##RGB color; the keyword argument name must be a standard mpl colormap name.
            return plt.cm.get_cmap(name, n)

    if len(res_img.shape) < 3:
        res_img = np.zeros((*image.shape[:2], 3))
        res_img[:,:,0] = image / np.amax(image)
    else:
        res_img = image
    if ax is None:
        fig, ax = plt.subplots(1,1)

    # If color is based on another column
    if color in dataframe.columns:
        unique_els = dataframe[color].unique()

        cmap = get_cmap(len(unique_els), name=colormap)
        patches = []
        for i, c in enumerate(unique_els):
            tmp_df = dataframe[dataframe[color] == c]
            curr_col = cmap(i)
            curr_col_3 = np.array(curr_col)[:3]

            for tmp_row in tmp_df.itertuples():
                tmp_row = tmp_row._asdict()
                if isinstance(dotsize, str):
                    tmp_dotsize = tmp_row[dotsize]
                else:
                    tmp_dotsize = dotsize

                rr, cc = disk((tmp_row[rowname], tmp_row[colname]), radius = tmp_dotsize, shape=res_img.shape)
                res_img[rr, cc, :] = curr_col_3

            patches.append(mpatches.Patch(color=curr_col_3, label=c))
        ax.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0. )

    else:
        tmp_color = colors.to_rgb(color)
        for tmp_row in dataframe.itertuples():
            tmp_row = tmp_row._asdict()

            if isinstance(dotsize, str):
                tmp_dotsize = tmp_row[dotsize]
            else:
                tmp_dotsize = dotsize

            rr, cc = disk((tmp_row[rowname], tmp_row[colname]), radius = tmp_dotsize, shape=res_img.shape)
            res_img[rr, cc, :] = tmp_color
        # ax.scatter(dataframe[colname], dataframe[rowname], color=color, s=tmp_dotsize)


    ax.imshow(res_img)

    if plot:
        plt.show()

    return ax, res_img

def plotObjectPerAreaBin(labeled_image, measured_df, mn, mx, interval):
    bins = list(range(mn, mx, interval))
    bins.append(float("inf"))

    str_labels = []
    for i, el in enumerate(bins):
        if not i == len(bins)-2:
            str_labels.append(f"{el}-{bins[i+1]}")
        else:
            str_labels.append(f"{bins[-2]}+")
            break
    print(str_labels)

    def get_cmap(n, name='tab10'):
        ##Returns a function that maps each index in 0, 1, ..., n-1 to a distinct 
        ##RGB color; the keyword argument name must be a standard mpl colormap name.
        return plt.cm.get_cmap(name, n)


    color_labels= []
    global_cmap = get_cmap(len(bins))

    for i in range(len(bins)-1):
        color_labels.append(global_cmap(i))

    area_interval_colors = pd.cut(measured_df["Area"], bins=bins, labels=color_labels)
    area_interval_strings = pd.cut(measured_df["Area"], bins=bins, labels=str_labels)
    measured_df["area_interval_colors"] = area_interval_colors
    measured_df["area_interval_strings"] = area_interval_strings

    plt.imshow(label2rgb(labeled_image, colors=measured_df["area_interval_colors"]))
    colors = color_labels
    # create a patch (proxy artist) for every color 
    patches = [ mpatches.Patch(color=colors[i], label=str_labels[i]) for i in range(len(str_labels)) ]
    plt.axis("off")
    # put those patched as legend-handles into the legend
    plt.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0. )







import os
import numpy as np
import SimpleITK as sitk
from skimage import io

def translationTranform(fixed, moving):
    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()
    R.SetOptimizerAsRegularStepGradientDescent(4.0, .01, 200)
    R.SetInitialTransform(sitk.TranslationTransform(fixed.GetDimension()))
    R.SetInterpolator(sitk.sitkLinear)
    outTx = R.Execute(fixed, moving)
    return outTx

def rigidTransform(fixed, moving):
    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()
    R.SetOptimizerAsRegularStepGradientDescent(4.0, .01, 200)
    initial_transform = sitk.CenteredTransformInitializer(
                        fixed,
                        moving,
                        sitk.Euler2DTransform(),
                        sitk.CenteredTransformInitializerFilter.GEOMETRY,
                    )
    R.SetInitialTransform(initial_transform)
    R.SetInterpolator(sitk.sitkLinear)
    outTx = R.Execute(fixed, moving)
    return outTx

def scaleTransform(fixed, moving):
    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()
    R.SetOptimizerAsRegularStepGradientDescent(4.0, .01, 200)
    initial_transform = sitk.CenteredTransformInitializer(
                        fixed,
                        moving,
                        sitk.ScaleTransform(fixed.GetDimension()),
                        sitk.CenteredTransformInitializerFilter.GEOMETRY,
                    )
    R.SetInitialTransform(initial_transform)
    R.SetInterpolator(sitk.sitkLinear)
    outTx = R.Execute(fixed, moving)
    return outTx

def multScaleTransform(fixed, moving, initial_transform):
    # This is the registration configuration which we use in all cases. The only parameter that we vary
# is the initial_transform.
    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
    R.SetMetricSamplingStrategy(R.RANDOM)
    R.SetMetricSamplingPercentage(0.01)
    R.SetInterpolator(sitk.sitkLinear)
    R.SetOptimizerAsGradientDescent(
        learningRate=1.0,
        numberOfIterations=100,
        estimateLearningRate=R.Once,
    )
    R.SetOptimizerScalesFromPhysicalShift()
    R.SetInitialTransform(initial_transform)
    R.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
    R.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
    R.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()
    outTx = R.Execute(fixed, moving)
    return outTx


def affineTransform(fixed, moving):
    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()
    R.SetOptimizerAsRegularStepGradientDescent(4.0, .01, 200)
    initial_transform = sitk.CenteredTransformInitializer(
                        fixed,
                        moving,
                        sitk.AffineTransform(fixed.GetDimension()),
                        sitk.CenteredTransformInitializerFilter.GEOMETRY,
                    )
    R.SetInitialTransform(initial_transform)
    R.SetInterpolator(sitk.sitkLinear)
    outTx = R.Execute(fixed, moving)
    return outTx

def bSplineTransform(fixed, moving):
    transformDomainMeshSize = [8] * moving.GetDimension()
    tx = sitk.BSplineTransformInitializer(fixed,
                                          transformDomainMeshSize)

    R = sitk.ImageRegistrationMethod()
    R.SetMetricAsCorrelation()

    R.SetOptimizerAsLBFGSB(gradientConvergenceTolerance=1e-5,
                           numberOfIterations=100,
                           maximumNumberOfCorrections=5,
                           maximumNumberOfFunctionEvaluations=1000,
                           costFunctionConvergenceFactor=1e+7)
    R.SetInitialTransform(tx, True)
    R.SetInterpolator(sitk.sitkLinear)

    outTx = R.Execute(fixed, moving)

    return outTx

def warpImage(image, transform: str or sitk.Transform) -> sitk.Image :
    if isinstance(transform, str):
        transform = sitk.ReadTransform(transform)
    resampled = sitk.Resample(image, transform, sitk.sitkLinear, 0.0, sitk.sitkUInt16)
    return resampled

def performRegistration(ref_image: str or sitk.Image or np.ndarray, target_image: str or sitk.Image or np.ndarray, method="rigid",out_name: str = "") -> np.array:
    transform_dict = {"translation": translationTranform, "rigid": rigidTransform, "affine": affineTransform, "bspline": bSplineTransform, "scale": scaleTransform}

    transformFunc = transform_dict[method]

    if isinstance(ref_image, str):
        ref = sitk.ReadImage(ref_image, sitk.sitkFloat32)
    elif isinstance(ref_image, np.ndarray):
        ref = sitk.GetImageFromArray(ref_image)
        ref = sitk.Cast(ref, sitk.sitkFloat32)

    else:
        ref = ref_image

    if isinstance(target_image, str):
        target = sitk.ReadImage(target_image, sitk.sitkFloat32)
    elif isinstance(target_image, np.ndarray):
        target = sitk.GetImageFromArray(target_image)
        target = sitk.Cast(target, sitk.sitkFloat32)
    else:
        target = target_image


    transform = transformFunc(ref, target)

    reformed_target = warpImage(target, transform)

    # if theres no out name but also no path of the target, we can't make an outname, so ti will not be saved
    if not isinstance(target_image, str) and not out_name:
        print("No out_name was given but target image is an image, so no output filepath can be generated. Registered image is not written to file.")
        reformed_target = sitk.GetArrayFromImage(reformed_target)
        return reformed_target

    # if the target was a string, but no outname was given, we have to derive the out_name first
    elif not out_name:
        out_name = os.path.splitext(target_image)[0]+ f"_{method}registered.tif"

    # finally we write the image and return it
    sitk.WriteImage(reformed_target, out_name)

    reformed_target = sitk.GetArrayFromImage(reformed_target)
    return reformed_target, out_name



import os
import cv2 as cv
import numpy as np
import pandas as pd
from skimage import io
import SimpleITK as sitk
import matplotlib.pyplot as plt
from skimage.color import rgb2gray
from skimage.transform import rotate
from .decorators import measureTime
from .measure import measureLabeledImage
from .manipulate import equalizeImageSize
from .register import performRegistration
from .tiling import tileGrid, Tile
from skimage.util import invert,img_as_ubyte, img_as_uint


def expandBbox(bbox, size: int):
    min_row, min_col, max_row, max_col = bbox
    min_row -= size
    max_row += size
    min_col -= size
    max_col += size
    bbox = min_row, min_col, max_row, max_col
    return bbox

def cutBboxFromImage(bbox, image: np.ndarray):
        min_row, min_col, max_row, max_col = bbox

        cut_image = image[min_row:max_row, min_col:max_col]
        #io.imsave(filename, cut_image)
        return cut_image

def isolateForeground(input_image: np.array, kernel_size:int = 100, bbox_expansion:int= 1000):
    # create labeled image
    # gray_correct = np.array(255 * (input_image / 255) ** 1.2 , dtype='uint8')
    gray_correct = img_as_ubyte(input_image)
    # Local adaptative threshold

    thresh = cv.adaptiveThreshold(gray_correct, 255, cv.ADAPTIVE_THRESH_GAUSSIAN_C, cv.THRESH_BINARY, 255, 19)
    thresh = cv.bitwise_not(thresh)

    # # Dilatation et erosion
    kernel = np.ones((kernel_size,kernel_size), np.uint8)
    img_dilation = cv.dilate(thresh, kernel, iterations=1)
    img_erode = cv.erode(img_dilation,kernel, iterations=1)
    # # clean all noise after dilatation and erosion
    img_erode = cv.medianBlur(img_erode, 7)

    ret, labeled_image = cv.connectedComponents(img_erode)


    # io.imsave("tmp_labeledimg.tif", labeled_image)
    # Meaure imageprops
    df = measureLabeledImage(labeled_image)

    # get max area row
    max_area_label_number = df.iloc[df['Area'].idxmax()]['image_label']

    # check if the object is horizontal or vertical, horizontal if orientation < 0
    bool_rotate = df.iloc[df['Area'].idxmax()]['orientation']

    # Set everything that isnt max label to zero
    labeled_image[labeled_image!=max_area_label_number] = 0
    image_bbox =  df.iloc[df['Area'].idxmax()]['bbox']
    image_bbox = expandBbox(image_bbox, bbox_expansion)

    foreground_image = cutBboxFromImage(image_bbox, input_image)

    return foreground_image, bool_rotate

def ST_to_MSI_registration(fixed_path: str, moving_path:str):
    """ST_to_MSI_registration.

    BE WARNED. This function is CPU and Memory intensive!

    """

    # fixed_image = sitk.ReadImage(fixed_path, sitk.sitkFloat32)
    # moving_image = sitk.ReadImage(moving_path, sitk.sitkFloat32)

    out_file_base = os.path.splitext(moving_path)[0]

    ref_image = io.imread(fixed_path)
    target_image = io.imread(moving_path)

    ref_image = invert(ref_image)

    target_image = invert(target_image)


    # Crop foreground from the weird Spatial H&E stain
    target_image, bool_rotate = isolateForeground(target_image, bbox_expansion= 500)

    if bool_rotate:
        target_image = rotate(target_image, -90, resize=True)

    # Downscale
    target_image = equalizeImageSize(ref_image, target_image)

    # Equalize ref image dtype, cause target image comes out as uint
    ref_image = img_as_uint(ref_image)

    fixed_image = sitk.GetImageFromArray(ref_image)

    moving_image = sitk.GetImageFromArray(target_image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)
    
    initial_transform = sitk.CenteredTransformInitializer(
    fixed_image,
    moving_image,
    sitk.AffineTransform(2),
    sitk.CenteredTransformInitializerFilter.GEOMETRY,
    )

    moving_resampled = sitk.Resample(
        moving_image,
        fixed_image,
        initial_transform,
        sitk.sitkLinear,
        0.0,
        moving_image.GetPixelID(),
    )
    registration_method = sitk.ImageRegistrationMethod()

    # Similarity metric settings.
    registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=50)
    registration_method.SetMetricSamplingStrategy(registration_method.RANDOM)
    registration_method.SetMetricSamplingPercentage(0.10)

    registration_method.SetInterpolator(sitk.sitkLinear)

    # Optimizer settings.
    registration_method.SetOptimizerAsGradientDescent(
        learningRate=1.0,
        numberOfIterations=50000,
        convergenceMinimumValue=1e-6,
        convergenceWindowSize=10,
    )
    registration_method.SetOptimizerScalesFromPhysicalShift()

    # Setup for the multi-resolution framework.
    registration_method.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
    registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
    registration_method.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    # Don't optimize in-place, we would possibly like to run this cell multiple times.
    registration_method.SetInitialTransform(initial_transform)

    intermediate_transform = registration_method.Execute(
        sitk.Cast(fixed_image, sitk.sitkFloat32), sitk.Cast(moving_image, sitk.sitkFloat32)
    )
    print(f"Affine Final metric value: {registration_method.GetMetricValue()}")
    print(f"Affine Optimizer's stopping condition, {registration_method.GetOptimizerStopConditionDescription()}")


    bspline_registration_method = sitk.ImageRegistrationMethod()

    transformDomainMeshSize = [8] * moving_resampled.GetDimension()
    tx = sitk.BSplineTransformInitializer(fixed_image, transformDomainMeshSize)


    bspline_registration_method.SetMetricAsCorrelation()

    bspline_registration_method.SetOptimizerAsLBFGSB(gradientConvergenceTolerance=1e-5,
                               numberOfIterations=100,
                               maximumNumberOfCorrections=5,
                               maximumNumberOfFunctionEvaluations=1000,
                               costFunctionConvergenceFactor=1e+7)
    bspline_registration_method.SetInitialTransform(tx, inPlace=False)
    bspline_registration_method.SetInterpolator(sitk.sitkLinear)

    #bspline_registration_method.SetOptimizerScalesFromPhysicalShift()

    # Setup for the multi-resolution framework.
    #bspline_registration_method.SetShrinkFactorsPerLevel(shrinkFactors=[4, 2, 1])
    #bspline_registration_method.SetSmoothingSigmasPerLevel(smoothingSigmas=[2, 1, 0])
    #bspline_registration_method.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()


    final_transform = bspline_registration_method.Execute(
        sitk.Cast(fixed_image, sitk.sitkFloat32), sitk.Cast(moving_resampled, sitk.sitkFloat32)
    )
    print(f"Bspline Final metric value: {bspline_registration_method.GetMetricValue()}")
    print(f"Bspline Optimizer's stopping condition, {bspline_registration_method.GetOptimizerStopConditionDescription()}")

    moving_bspline_resampled = sitk.Resample(
        moving_resampled,
        fixed_image,
        final_transform,
        sitk.sitkLinear,
        0.0,
        moving_image.GetPixelID(),
    )
    sitk.WriteImage(
        moving_bspline_resampled, os.path.join(f"{out_file_base}_registered.tif")
    )
    sitk.WriteTransform(
        final_transform, os.path.join(f"{out_file_base}_transform.txt")
    )

    return moving_bspline_resampled, final_transform



def assignGenesToAllTiles(image_base_name, measured_base_name, detected_genes_base_name, grid: tileGrid, out_dir = "./"):
    '''
    base name being something tile merfish_registered_tile
    '''
    for i in range(1, grid.n_tiles + 1):
        img = io.imread(f"{image_base_name}{i}.tif")
        df = pd.read_csv(f"{detected_genes_base_name}{i}.csv")
        measured_df = pd.read_csv(f"{measured_base_name}{i}.csv")
        tile = Tile(img, df, measured_df, grid, i)

        try:
            img1 = io.imread(f"{image_base_name}{i-1}.tif")
            df1 = pd.read_csv(f"{detected_genes_base_name}{i-1}.csv")
            measured_df1 = pd.read_csv(f"{measured_base_name}{i-1}.csv")
            tile1 = Tile(img1, df1, measured_df1, grid, i-1)
        except:
            tile1 = None

        try:
            img2 = io.imread(f"{image_base_name}{i- tile.coldiv}.tif")
            df2 = pd.read_csv(f"{detected_genes_base_name}{i-tile.coldiv}.csv")
            measured_df2 = pd.read_csv(f"{measured_base_name}{i-tile.coldiv}.csv")
            tile2 = Tile(img2, df2, measured_df2, grid2, i- tile.coldiv)
        except:
            tile2 = None

        count_matrix = tile.createCountMatrix(tile1, tile2)
        count_matrix.to_csv(os.path.join(out_dir, f"{os.path.basename(image_base_name)}{i}_countmatrix.csv"))



import pandas as pd
import numpy as np
from skimage import measure

def measureLabeledImage(labeled_image: np.array, original_image: np.array = None, pixels_to_um:int = 0) -> pd.DataFrame:
    regions = measure.regionprops(labeled_image, intensity_image=original_image)

    propList = ['Area',
                'bbox',
                'equivalent_diameter', 
                'orientation', 
                'MajorAxisLength',
                'MinorAxisLength',
                'Perimeter',
                'MinIntensity',
                'MeanIntensity',
                'MaxIntensity']    

    rows_list=[]
    for region_props in regions:
        attribute_dict = {}
        center_row, center_col= region_props['centroid']
        attribute_dict['image_label'] =region_props['Label']
        attribute_dict['cell_label'] = f"row{int(center_row)}_col{int(center_col)}_{region_props['Label']}"
        attribute_dict['center_row'] = int(center_row)
        attribute_dict['center_col'] = int(center_col)
        for i,prop in enumerate(propList):
            if(prop == 'Area') and pixels_to_um != 0: 
                attribute_dict['real_area'] = region_props[prop]*pixels_to_um**2
            elif (prop.find('Intensity') < 0) and pixels_to_um != 0:          # Any prop without Intensity in its name
                attribute_dict[prop] = region_props[prop]*pixels_to_um
            elif (prop.find('Intensity') < 0):
                attribute_dict[prop] = region_props[prop]
            else: 
                if original_image is not None:
                    attribute_dict[prop] = region_props[prop]

        rows_list.append(attribute_dict)
    attribute_df = pd.DataFrame(rows_list)
    return attribute_df




import math
import re
import numpy as np
from pathlib import Path
from skimage import io
import pandas as pd
from scipy.ndimage import distance_transform_edt
from glob import glob


def getPatch(arr, idx, radius=3, fill=None):
    """
    Gets surrounding elements from a numpy array

    Parameters:
    arr (ndarray of rank N): Input array
    idx (N-Dimensional Index): The index at which to get surrounding elements. If None is specified for a particular axis,
        the entire axis is returned.
    radius (array-like of rank N or scalar): The radius across each axis. If None is specified for a particular axis,
        the entire axis is returned.
    fill (scalar or None): The value to fill the array for indices that are out-of-bounds.
        If value is None, only the surrounding indices that are within the original array are returned.

    Returns:
    ndarray: The surrounding elements at the specified index
    """

    assert len(idx) == len(arr.shape)

    if np.isscalar(radius):
        radius = tuple([radius for i in range(len(arr.shape))])

    slices = []
    paddings = []
    for axis in range(len(arr.shape)):
        if idx[axis] is None or radius[axis] is None:
            slices.append(slice(0, arr.shape[axis]))
            paddings.append((0, 0))
            continue

        r = radius[axis]
        l = idx[axis] - r
        r = idx[axis] + r

        pl = 0 if l > 0 else abs(l)
        pr = 0 if r < arr.shape[axis] else r - arr.shape[axis] + 1

        slices.append(slice(max(0, l), min(arr.shape[axis], r+1)))
        paddings.append((pl, pr))

    if fill is None:
        return arr[tuple(slices)]
    return np.pad(arr[tuple(slices)], paddings, 'constant', constant_values=fill)

def getLastDimensions(array, nr_dims: int = 2):
    """Extracts last X dimensions from an array

    Parameters
    ----------
    array :
        input array to be extracted
    nr_dims : int
        nr of dimensions to extract. Default = 2 
    """
    if array.ndim <= nr_dims:
        return array
    if nr_dims > nr_dims:
        raise ValueError(f"Requested nr of dimensions {nr_dims} higher than nr of dimensions present in array.")
    slc = [0] * (array.ndim - nr_dims)
    slc += [slice(None) for i in range(0, nr_dims)]
    return array[tuple(slc)]

#TODO check userfulness
def isolateSingleCellsFromTile(measured_df_basename: str , image_basename: str, grid: "tileGrid", tile_nr, out_dir: str or Path = None, image_prefix: str = "", labeled_image_basename = None, GEimage_basename = None):

    # Input parsing
    if image_prefix and not image_prefix.endswith("_"):
        image_prefix += "_"

    if not out_dir:
        out_dir = Path(".")
    else:
        out_dir = Path(out_dir)

    out_dir.mkdir(exist_ok=True)

    if GEimage_basename is not None:
        GE_out_dir = out_dir / "gene_expression_imgs/"
        GE_out_dir.mkdir(exist_ok=True)

    regular_dir = out_dir / "regular_bbox/"  
    regular_dir.mkdir(exist_ok=True)

    # first we build the image
    middle_tile = io.imread(f"{image_basename}{tile_nr}.tif")
    measured_df = pd.read_csv(f"{measured_basename}{tile_nr}.csv")

    neighbours = grid.getNeighbouringTiles(tile_nr)

    all_4_shape =(grid.tile_size_row * 3, grid.tile_size_col * 3) # If all neighbours exist
    empty = np.zeros(all_4_shape)

    # create list of neighbours such that looping is ordered 
    # Now fill in the images in the large empty image
    row_size = grid.tile_size_row
    col_size = grid.tile_size_col
    if neighbours["top"] is not None:
        top_neighbour =  io.imread(f"{image_basename}{neighbours['top']}.tif")
        top_measured_df = pd.read_csv(f"{measured_basename}{neighbours['top']}.csv")
        empty[0:row_size, col_size: 2*col_size] = top_neighbour

    if neighbours["left"] is not None:
        left_neighbour =  io.imread(f"{image_basename}{neighbours['left']}.tif")
        left_measured_df = pd.read_csv(f"{measured_basename}{neighbours['left']}.csv")
        empty[row_size : 2*row_size, 0:col_size] =left_neighbour

    if neighbours["right"] is not None:
        right_neighbour =  io.imread(f"{image_basename}{neighbours['right']}.tif")
        right_measured_df = pd.read_csv(f"{measured_basename}{neighbours['right']}.csv")
        empty[row_size : 2*row_size, 2*col_size : 3*col_size] =right_neighbour

    if neighbours["bot"] is not None:
        bot_neighbour =  io.imread(f"{image_basename}{neighbours['bot']}.tif")
        bot_measured_df = pd.read_csv(f"{measured_basename}{neighbours['bot']}.csv")
        empty[2*row_size : 3*row_size, col_size : 2*col_size] = bot_neighbour

    # Now add the middle tile as well
    empty[row_size : 2*row_size, col_size : 2*col_size] = middle_tile


    # then we adapt the center locations in the context of our custom 3x3 grid 
    def adaptCoordsOfMiddleTile(row, col, grid):
        new_row = row + grid.tile_size_row
        new_col = col + grid.tile_size_col

        return new_row, new_col
    def adaptCoordsOfBottomTile(row, col, grid):
        new_row = row + 2*grid.tile_size_row
        new_col = col + grid.tile_size_col

        return new_row, new_col
    def adaptCoordsOfRightTile(row, col, grid):
        new_row = row + grid.tile_size_row
        new_col = col + 2*grid.tile_size_col

        return new_row, new_col

    plt.imshow(empty)
    # plt.show()
    # Check which labels shouldn't be included in this iteration of isolation (so those bot and right)
    middle_border = tileBorder(middle_tile, tile_nr)

    ## Take the list of the labels of objects that are located on the top and left border 
    labels_to_leave_out = (*middle_border.orientation_label_centers["top"].values(), *middle_border.orientation_label_centers["left"].values())
    labels_on_the_border =  (*middle_border.orientation_label_centers["bot"].values(), *middle_border.orientation_label_centers["right"].values())

    #TODO last step: align centers for new_row, new_col, it doesn't yet because the cells on the border's center of gravity is incorrect when measured on just the tile
    # Maybe here we take the corresponding label and just calculate the middle of the 2
    # tile = tileBorder(img, i)
    # try:
    #     img2 =  io.imread(labeled_image_paths[i])
    #     tile2 = tileBorder(img2, i+1)
    #     tile2_matches = tile.matchTileBorders(tile2, tile_grid)
    def calcMidpointByArea(point1, point2, point1_ratio, point2_ratio):
        assert point1_ratio + point2_ratio == 1.0
        return ((point1[0] + point2[0])/2, (point1[1] + point2[1])/2)
    def weighted_midpoint(center1, area1, center2, area2):
        weighted_center1 = tuple(x * area1 for x in center1)
        weighted_center2 = tuple(x * area2 for x in center2)
        total_area = area1 + area2
        weighted_midpoint = tuple((x + y) / total_area for x, y in zip(weighted_center1, weighted_center2))
        return weighted_midpoint

     ## # then we can loop over the df and extract the images from the enlarged image
    for i, cell_row in enumerate(measured_df.itertuples()):
        if cell_row.image_label not in labels_to_leave_out:
            if cell_row.image_label in labels_on_the_border:
                row = cell_row.center_Y
                col = cell_row.center_X
                new_row, new_col = adaptCoordsOfMiddleTile(row, col, grid)
                # If this cell is on the border, it's center is wrong, so we find the corresponding label on the neighbouring tile and calculate the mean distance between the two as new center of the mosaic
                if cell_row.image_label in middle_border.orientation_label_centers["bot"].values(): # if the label is on the bottom we need to look at bottom label
                    # first create border of neighbour and find matching label
                    bot_border = tileBorder(bot_neighbour, neighbours["bot"])
                    bot_matches = middle_border.matchTileBorders(bot_border, grid)
                    matching_label = bot_matches[cell_row.image_label]
                    print(bot_measured_df.loc[bot_measured_df['image_label'] == matching_label]['center_X'])
                    matching_label_center_row, matching_label_center_col = bot_measured_df.loc[bot_measured_df['image_label'] == matching_label]['center_Y'][0],  bot_measured_df.loc[bot_measured_df['image_label'] == matching_label]['center_X'][0]
                    # then adapt that label's coordinates to the current 3x3 coordinate system
                    new_matching_label_center_row, new_matching_label_center_col  = adaptCoordsOfBottomTile(matching_label_center_row, matching_label_center_col, grid)

                    # Then define a new point in between this and matching center, weighted by their areas
                    current_point_area = cell_row.Area
                    matching_point_area = bot_measured_df.loc[bot_measured_df['image_label'] == matching_label]['Area'][0]
                    midpoint_row, midpoint_col = weighted_midpoint((new_matching_label_center_row, new_matching_label_center_col), matching_point_area, (new_row,new_col), current_point_area) 
                    plt.scatter((new_col, new_matching_label_center_col, midpoint_col), (new_row, new_matching_label_center_row, midpoint_row), color=('red', 'green', 'orange'))
                    # plt.show()

            else:
                pass
                # row = cell_row.center_Y
                # col = cell_row.center_X
                # new_row, new_col = adaptCoordsOfMiddleTile(row, col, grid)
                # bb = getPatch(empty, (new_row,new_col), radius = (149,149), fill=0)
                # io.imsave(regular_dir / f"{image_prefix}cell{i}.tif", bb)
                # if labeled_image is not None:
                #     labeled_bb = getPatch(labeled_image, (Y,X), radius = (74,74), fill=0)
                #     bb[labeled_bb != cell_row.image_label] = 0 
                # io.imsave(out_dir / f"{image_prefix}cell{i}.tif", bb)

                # if GEimage is not None:
                #     gene_labeled_bb = getPatch(GEimage, (Y,X), radius = (74,74), fill=0)
                #     gene_labeled_bb[labeled_bb != cell_row.image_label] = 0
                #     io.imsave(GE_out_dir / f"{image_prefix}gene_labeled{i}.tif", gene_labeled_bb)
                # plt.imshow(bb)
                # plt.title(f"tile {tile_nr}, label: {cell_row.image_label}")
            # plt.gca().invert_yaxis()
            # plt.scatter((col, new_col), (row, new_row), color=('red', 'green'))
            # plt.show()
    # plt.show()

def sortByRegex(target_list, regex):
    r = re.compile(rf"{regex}")
    def key_func(m):
        return int(r.search(m).group(1))

    target_list.sort(key=key_func)
    return target_list

def findCenter(x_shape):
    middle = x_shape[0] / 2, x_shape[1] / 2
    return middle

def exceptGlob(full_glob: str, except_glob: str):
    return list(set(glob(full_glob)) - set(glob(except_glob)))

def expand_labels_manual(label_image, distance=1):
    distances, nearest_label_coords = distance_transform_edt(
        label_image == 0, return_indices=True
    )
    labels_out = np.zeros_like(label_image)
    dilate_mask = distances <= distance

    masked_nearest_label_coords = [
        dimension_indices[dilate_mask]
        for dimension_indices in nearest_label_coords
    ]
    nearest_labels = label_image[tuple(masked_nearest_label_coords)]
    labels_out[dilate_mask] = nearest_labels
    return labels_out, distances, nearest_label_coords

def expand_labels_on_something_else(label_image, distances, nearest_label_coords, distance=1):
    labels_out = np.zeros_like(label_image)
    dilate_mask = distances <= distance
    masked_nearest_label_coords = [
        dimension_indices[dilate_mask]
        for dimension_indices in nearest_label_coords
    ]
    nearest_labels = label_image[tuple(masked_nearest_label_coords)]
    labels_out[dilate_mask] = nearest_labels
    return labels_out

if __name__ == '__main__':
    import pickle
    import matplotlib.pyplot as plt
    from tiling import tileGrid, tileBorder

    image_basename = "../out_dir/labeled1_MERFISH_nuclei_tile"
    measured_basename = "../out_dir/labeled1_MERFISH_nuclei_measured_tile"
    with open("../out_dir/tile_grid.pickle", 'rb') as f:
        grid = pickle.load(f)
    # grid = tileGrid(4,4,2048, 2048)
    for i in range(1,17):
        print(f"{i} iteration")
        isolateSingleCellsFromTile(measured_basename, image_basename, grid, i, out_dir = "../out_dir/isolated_images/")
        



import os
from glob import glob
import re
import numpy as np
import pandas as pd
from skimage import io
from tifffile import imsave
from typing import List, Tuple
import matplotlib.pyplot as plt
from skimage.io import imread_collection, imsave, imread
import PIL
PIL.Image.MAX_IMAGE_PIXELS = None



def calculateOptimalLargestResolution(glob_pattern: str, target_tile_height: int, target_tile_width: int) -> Tuple[int]: 
    """Calculates the optimal maximum resolution based on a set of images assuming that it will be used to tile the images into equal tiles of a certain size, given by the input variables.


    Parameters
    ----------
    glob_pattern : str
        glob pattern hitting all images to be taken into account for the calculation.
    target_tile_height : int
        target_tile_height
    target_tile_width : int
        target_tile_width

    Returns
    -------
    Tuple[int]

    """
    
    def calcMaxesOfNonLoadedImages(str_glob_pattern):
        heights = []
        widths = []
        for file in glob(str_glob_pattern):
            with PIL.Image.open(file) as img:
                width, height = img.size         
                widths.append(width)
                heights.append(height)
        return max(heights), max(widths)

    def calcMaxes(images_array):
        heights = []
        widths = []
        if len( images_array.shape ) > 2:
            for i in range(len(images_array)):
                heights.append(images_array[i].shape[0])
                widths.append(images_array[i].shape[1])
            max_rows = max(heights) 
            max_columns = max(widths)
        else:
            max_rows=images_array.shape[0]
            max_columns=images_array.shape[1]
        return max_rows, max_columns

    if isinstance(glob_pattern, str):
        # images_array = np.array(imread_collection(glob_pattern))
        max_rows, max_columns = calcMaxesOfNonLoadedImages(glob_pattern)
    elif isinstance(glob_pattern, np.ndarray):
        images_array = glob_pattern
        max_rows, max_columns = calcMaxes(images_array)
    elif isinstance(glob_pattern, tuple):
        # If the given argument is just already the max resolution
        max_rows, max_columns = glob_pattern

    rowdiv = np.ceil(max_rows / target_tile_height)
    coldiv = np.ceil(max_columns / target_tile_width)

    target_full_rows = rowdiv * target_tile_height
    target_full_columns = coldiv * target_tile_width

    return target_full_rows, target_full_columns, rowdiv, coldiv

def padImage(image: np.array, target_full_rows: int, target_full_columns: int) -> np.ndarray:
    """Pads an image with black pixels up until a certain number of rows and columns

    Parameters
    ----------
    image : np.array
        image to be padded
    target_full_rows : int
        Number of rows to be achieved by padding
    target_full_columns : int
        Number of columns to be achieved by padding

    Returns
    -------
    np.ndarray

    """
    rowdiff = target_full_rows - image.shape[0]
    columndiff = target_full_columns - image.shape[1]
    padded_img = np.pad(image, ((0, rowdiff), (0, columndiff)))
    return padded_img

def tileImage(image: np.ndarray, rowdiv: int, coldiv: int, image_prefix: str="image_tile_"):
    """Tile an image into smaller tiles based on divisions in x and y axes, and saving them to tif files.

    Parameters
    ----------
    image : np.ndarray
        image
    rowdiv : int
        Nr of divisions to make in the y axis
    coldiv : int
        Nr of divisions to make in the x axis
    image_prefix : str
        prefix to add to the new filenames of the tiles. default = image_tile
    """
    if not image_prefix.endswith("_"):
        image_prefix += "_"

    temp_split = np.array_split(image, rowdiv, axis = 0)
    # Item sublist part is just to unpack a list of lists into one list
    final_split = [item for sublist in [np.array_split(row, coldiv, axis = 1) for row in temp_split] for item in sublist]

    for i, img in enumerate(final_split, 1):
        imsave(f"{image_prefix}tile{i}.tif", img, check_contrast=False)

def tile(glob_pattern: str, target_tile_width: int, target_tile_height: int, out_dir: str = "", calc_only=False) -> Tuple[int]:
    """Tile the images caught by the glob pattern by first padding them to a global image size needed to tile them all into the same tile size, given by the input values. Tiles are written to tif files in the basedir of the original image, or to out_dir if given, with naming convention {out_dir | basedir}/{image_name}_tile{i}.tif

    Parameters
    ----------
    glob_pattern : str
        glob_pattern that catches all images to be tiled.
    target_tile_width : int
        Width of the tiles
    target_tile_height : int
        Height of the tiles
    calc_only: bool
        If true, only calculates the padded image size and number of tiles in both dimensions, without creating the tiles. 
        Output can be used to create a tileGrid object. 
        

    Returns
    -------
    Tuple[int]

    """
    if out_dir and not calc_only:
        os.makedirs(out_dir, exist_ok=True)

    target_full_rows, target_full_columns, rowdiv, coldiv = calculateOptimalLargestResolution(glob_pattern, target_tile_height, target_tile_width)

    if calc_only:
        grid = tileGrid(rowdiv, coldiv, target_full_rows, target_full_columns)
        return grid

    padded_imgs = {}
    if not out_dir:
        for image_path in glob(glob_pattern):
            padded_imgs[os.path.splitext(image_path)[0]] = padImage(imread(image_path), int(target_full_rows), int(target_full_columns))
    else:
        for image_path in glob(glob_pattern):
            padded_imgs[os.path.join(out_dir, os.path.splitext(os.path.basename(image_path))[0])] = padImage(imread(image_path), int(target_full_rows), int(target_full_columns))

    for k, padded_img in padded_imgs.items():
        tileImage(padded_img,rowdiv = rowdiv, coldiv = coldiv, image_prefix = k)
    grid = tileGrid(rowdiv, coldiv, target_full_rows, target_full_columns)
    return rowdiv,coldiv, target_full_rows, target_full_columns

class tileGrid:
    # """ Class to represent the tiling grid of the tile function
    # Important for usage: tile names start at indexin 1, so for self.tile_boundaries (which is a dict), keys start at 1
    # image list and data coordinates are there to be able to tile them after the fact
    # """
    def __init__(self, rowdiv, coldiv, n_rows, n_cols, image_list = [], data_coordinates_list=[], tiles={}, pattern="row_by_row"):
# basic vars
        self.n_rows = n_rows # Number rows in the complete array, untiled
        self.n_cols = n_cols # Number cols in the complete array, untiled
        self.coldiv = int( coldiv ) # Number of tiles in the col-dimension
        self.rowdiv = int( rowdiv ) # Number of tiles in the row dimension
        self.image_list = image_list # List of images of the original image-stack ,in case the tiling tiles more than 1 image in unison
        self.data_coordinates_list = data_coordinates_list # List of potential dataframes with point coordinates, such as spatial transcriptomics
        self.tiles = tiles

        # calculated stuff
        self.tile_size_row =int(  n_rows / rowdiv ) # Size of each individual tile in row dimension
        self.tile_size_col =int(  n_cols / coldiv ) # Size of each individual tile in col dimension
        self.n_tiles = int(rowdiv * coldiv) # total number of tiles created

        # numerical representation of the tiles
        if pattern == "row_by_row":
            self.tile_grid = np.arange(1, self.n_tiles + 1).reshape(self.rowdiv, self.coldiv) # representation of the tile grid in numbers, for if you want to search for the position of a specific tile
        elif pattern == "snake_by_row":
            '''
            [[ 1 2 3 ]
             [ 6 5 4 ]
             [ 7 8 9 ]]
            '''
            arr = np.arange(1, self.n_tiles+1)
            # Reshape the range into a 2D array
            arr_2d = arr.reshape(self.rowdiv, self.coldiv)
            # Reverse every other row
            for i in range(1, arr_2d.shape[0], 2):
                arr_2d[i] = arr_2d[i, ::-1]
            # Print the resulting array
            self.tile_grid = arr_2d

        # Calculate boundaries of tiles
        self.tile_boundaries = {} # Dict that stores the boundaries of the tiles (with respect to the padded image, not the original)
        # start at one to make the boundary math check out
        for i in range(1, self.n_tiles + 1):
            idx = np.where(self.tile_grid == i)
            self.tile_boundaries[i] = np.s_[idx[0][0] * self.tile_size_row : (idx[0][0] + 1) * self.tile_size_row, idx[1][0] * self.tile_size_col: (idx[1][0] + 1) * self.tile_size_col]

    def addImage(self, image: np.ndarray):
        self.image_list.append(image)

    def addTile(self, tile: "Tile"):
        self.tiles[tile.tile_nr] = tile

    def getTile(self, tile_nr):
        return self.tiles[tile_nr]

    def plotImageTile(self, tile_nr = 1, image_nr = 0):
        plt.imshow(self.image_list[image_nr][self.tile_boundaries[tile_nr]])
        plt.show()

    def getImageTile(self, tile_nr = 1, image_nr = 0):
        return self.image_list[image_nr][self.tile_boundaries[tile_nr]]

    def addDataCoordinates(self, data_df):
        self.data_coordinates_list.append(data_df)

    def getTileDataCoordinates(self, tile_nr, data_index=0,rowname="row", colname="col"):
        """getTileDataCoordinates.
        Note to self: because padding happens only at the end of the dimensions, it's not included in the tiling of coordinates, since their coordinate is the same relative to the padded image.

        Parameters
        ----------
        tile_nr :
            nr of tile to fetch coordinates from. (indexing starts at 1, not at 0)
        data_index :
            index of which dataframe to fetch, in case multiple dataframes are linked with the tiling coordinate. (indexing starts at 0)
        rowname :
            column name that refers to the row dimension
        colname :
            column name that refers to the col dimension
        """

        df = self.data_coordinates_list[data_index]
        cropped_df = self._cropCoordinateDf(df, tile_nr, rowname=rowname, colname=colname)

        return cropped_df

    def _cropCoordinateDf(self, df, tile_nr, rowname="row", colname="col"):
        row_slice, col_slice =  self.tile_boundaries[tile_nr][0], self.tile_boundaries[tile_nr][1]

        tmp_df = df.loc[(df[rowname] >= row_slice.start) & (df[rowname] < row_slice.stop)]
        cropped_df = tmp_df.loc[(tmp_df[colname] > col_slice.start) & (tmp_df[colname] <= col_slice.stop)]

        local_rows = [el - row_slice.start for el in cropped_df[rowname]]
        local_cols = [el - col_slice.start for el in cropped_df[colname]]

        # copy line added since assigning new columns to the normal cropped_df (which is technically a slice of the original df) raises a warning.
# In this usecase the warning is a false positive, since we don't care about tracing the new column back to the original dataframe
        cropped_df = cropped_df.copy()
        cropped_df["local_row"] = local_rows
        cropped_df["local_col"] = local_cols
        return cropped_df 

    def __str__(self):
        return f"Tile grid of size {self.rowdiv} by {self.coldiv}, {self.n_tiles} in total.\nTiles are {self.tile_size_row} rows by {self.tile_size_col} cols.\n {self.tile_grid}"

    def getNeighbouringTiles(self, tile_nr):
        flattened_grid = self.tile_grid.flatten()
        idx = np.argwhere(self.tile_grid == tile_nr)[0]
        left = tile_nr - 1 if (tile_nr - 1  in flattened_grid) and (idx[1] !=  0) else None
        right =  tile_nr + 1 if (tile_nr + 1 in flattened_grid) and (idx[1] != self.coldiv - 1) else None
        top =  tile_nr - self.coldiv if (tile_nr - self.coldiv in flattened_grid) and (idx[0] != 0) else None
        bot =  tile_nr + self.coldiv if (tile_nr + self.coldiv in flattened_grid) and (idx[0] != self.rowdiv-1) else None
        return {"left": left, "top": top, "bot": bot, "right": right}

class tileBorder:
    """ Class to represent the border of a tile. Requires the labeled image of that tile for initialization, since this class is meant to compare
    {orientation: {(row,col): label} }
    """
    def __init__(self, labeled_image:np.ndarray, tile_nr:int=0):
        self.labeled_image = labeled_image

        self.border_mask = self._getArrayBorder(self.labeled_image)
        tmp_list = list( np.unique(self.labeled_image[self.border_mask]))
        # Remove 0, which is background, not an actual label
        if 0 in tmp_list:
            tmp_list.remove(0)
        self.border_labels = tmp_list 
        self.nr_border_objects = len(self.border_labels)
        self.tile_nr = tile_nr

        ## only for in-class use
        self._barDict = {"left": self.labeled_image[:, 0], "top": self.labeled_image[0, :], "right" : self.labeled_image[:, -1], "bot": self.labeled_image[-1, :]}

        self.orientation_label_centers = self._getLabelCentersPerOrientation()

        if tile_nr != 0:
            self.tile_nr = tile_nr

    def _getLabelCentersPerOrientation(self):
        """ 
        Creates a dict represention of at which border is each label, in what its center is.

        returns: {orientation: {(row,col): label} }  with orientation either top, right, left of bot
        """ 
        orientation_dict = {}
        for orientation, bar  in self._barDict.items():
            labels = list(np.unique(bar))
            labels.remove(0)
            orientation_dict[orientation] = {}
            for label in labels:
                label_indexes = np.where(bar == label)[0]
                middle = label_indexes[round(len(label_indexes)/2)]
                orientation_dict[orientation][middle] = label
        return orientation_dict

    def _getArrayBorder(self,image):
        mask = np.ones(image.shape, dtype = bool)
        mask[image.ndim * (slice(1, -1), )] = False
        return mask

    def __eq__(self, other):
        return self.tile_nr == other.tile_nr

    def __lt__(self, other):
        return self.tile_nr < other.tile_nr

    def __gt__(self, other):
        return self.tile_nr > other.tile_nr

    def plotBorderCenters(self):
        plt.imshow(self.labeled_image)
        for k, v in self.orientation_label_centers["left"].items():
            plt.scatter(0,k)
        for k, v in self.orientation_label_centers["top"].items():
            plt.scatter(k,0)
        for k, v in self.orientation_label_centers["right"].items():
            plt.scatter((self.labeled_image).shape[1]-1,k)
        for k, v in self.orientation_label_centers["bot"].items():
            plt.scatter(k,(self.labeled_image).shape[0]-1)
        plt.show()

    def matchTileBorders(self, tileBorder2, tile_grid: "tileGrid", error_margin = 5):
        ## we take the largest number as viewpoint, which means it will be bordering the other tile 
        ref, other = (self, tileBorder2) if self > tileBorder2 else (tileBorder2, self)
        # print(ref.orientation_label_centers, other.orientation_label_centers)

        # if other is next to this one, look right to left
        if other.tile_nr == (ref.tile_nr - 1):
            ref_border = "left"
            other_border = "right"
        # if other is one row further, look bot to top
        elif other.tile_nr == (ref.tile_nr - tile_grid.coldiv):
            ref_border = "top"
            other_border = "bot"
        else:
            return {}

        labels_matching_dict = {} # this will store mapping from labels ref to other

        # get all centers of the ref
        center_dict = ref.orientation_label_centers[ref_border]
        if center_dict: # if there are any
            for ref_key in center_dict.keys():
                # for each one, check which centers there are on the other border border 
                other_key_array = np.array(list(other.orientation_label_centers[other_border].keys()))
                # build in an error margin since rounding errors are possible on the center detection
                matching_label = other.orientation_label_centers[other_border][other_key_array[np.where((other_key_array >= (ref_key - error_margin)) & (other_key_array <= (ref_key + error_margin)))][0]]
                labels_matching_dict[center_dict[ref_key]] = matching_label

        return labels_matching_dict

    def __str__(self):
        return f"Borders of tile {self.tile_nr}. {self.nr_border_objects} unique objects at borders."




class Tile:
    def __init__(self, labeled_image, detected_genes_df, measured_df, grid, tile_nr):
        self.labeled_image = labeled_image
        self.detected_genes_df = detected_genes_df
        self.grid = grid
        self.measured_df = measured_df
        self.tile_nr = tile_nr

    def getGenesOfLabel(self, label):
        gene_dict = {} # keys = label integers, values = dict{gene: count}
        gene_dict[label] = {}

        for row in self.detected_genes_df.itertuples():
            try:
                this_label = self.labeled_image[row.local_row, row.local_col]
            except IndexError:
                continue
            if label == this_label:
                gene_dict[label][row.Gene] = gene_dict[label].get(row.Gene, 0) + 1 
        return gene_dict

    
    def createCountMatrix(self, left_tile: "tileBorder" = None, top_tile: "tileBorder" = None):

        # utility function to concat the gene lists of two labels
        def concatGeneDicts(ref_gene_dict, target_gene_dict, ref_label, target_label):
            # iterate over target and add their counts to the ref
            for gene, count in target_gene_dict[target_label].items():
                ref_gene_dict[ref_label][gene] = ref_gene_dict[ref_label].get(gene, 0) + count 

        # Actual count matrix creation, including getting the counts from top and left
        def assignGenesToSpots(labeled_image, decoded_df):
            # First get all genes for this tile
            gene_dict = {} # keys = label integers, values = dict{gene: count}
            n_labels = np.unique(labeled_image)
            for label in n_labels:
                gene_dict[label] = {}
            for row in decoded_df.itertuples():
                try:
                    label = labeled_image[row.local_row, row.local_col]
                    if label != 0:
                        gene_dict[label][row.Gene] = gene_dict[label].get(row.Gene, 0) + 1 
                except:
                    pass

            this_border = self.getTileBorder()
            # then get genes of the tile to the left
            if left_tile is not None:
                left_border = left_tile.getTileBorder()
                left_matches = this_border.matchTileBorders(left_border, self.grid)

                for this_label, left_label in left_matches.items():
                    left_gene_dict = left_tile.getGenesOfLabel(left_label)
                    concatGeneDicts(gene_dict, left_gene_dict, this_label, left_label)


            if top_tile is not None:
                top_border = top_tile.getTileBorder()
                top_matches = this_border.matchTileBorders(top_border, self.grid)

                for this_label, top_label in top_matches.items():
                    top_gene_dict = top_tile.getGenesOfLabel(top_label)
                    concatGeneDicts(gene_dict, top_gene_dict, this_label, top_label)

            return gene_dict

        gene_dict = assignGenesToSpots(self.labeled_image, self.detected_genes_df)

        gene_dict_list = [gene_dict[i] for i in sorted(gene_dict.keys())]

        keys = set().union(*gene_dict_list)
        final = {k: [d.get(k, 0) for d in gene_dict_list] for k in keys}

        count_matrix = pd.DataFrame(final)

        return count_matrix


    def getTileBorder(self):
        return tileBorder(self.labeled_image, self.tile_nr)


def mergeLabeledTiles(labeled_image_paths, tile_grid):
    """
    This is used after segmenting tiles seperately since the whole dapi won't fit into memory.
    """
    from skimage.measure import label
    r = re.compile(r"tile(\d+)")
    def key_func(m):
        return int(r.search(m).group(1))
    labeled_image_paths.sort(key=key_func)


    empty = np.zeros((int(grid.n_rows), int(grid.n_cols)), np.uint)

    # enumerate starting from one cause tile counts start at 1
    for tile_path in labeled_image_paths:
        tile_nr = int(re.findall('\d+', re.findall('tile\d+', tile_path)[0])[0])
        tile = io.imread(tile_path)
        empty[grid.tile_boundaries[tile_nr]] = tile

#     convert all labels back into a mask so that we can label it again
    empty[empty != 0] = 1
    labeled_image = label(empty, background = 0)
    return labeled_image

def plotLabeledImageOverlap(labeled_image_paths, tile_grid):
    """Plots labeled images of a tile grid in the correct axis, with overlapping labels as their title
    Pure for visualization of the tileBorder class
    """

    r = re.compile(r"tile(\d+)")
    def key_func(m):
        return int(r.search(m).group(1))
    labeled_image_paths.sort(key=key_func)
    print(labeled_image_paths)

    fig, axs = plt.subplots(tile_grid.rowdiv, tile_grid.coldiv)
    axs = axs.flatten()
    for i in range(1,tile_grid.n_tiles + 1):
        img = io.imread(labeled_image_paths[i-1])
    
        tile = tileBorder(img, i)
        try:
            img2 =  io.imread(labeled_image_paths[i])
            tile2 = tileBorder(img2, i+1)
            tile2_matches = tile.matchTileBorders(tile2, tile_grid)
        except:
            tile2_matches = {}
        try:
            img3 =  io.imread(labeled_image_paths[i + tile_grid.coldiv - 1])
            tile3 = tileBorder(img3, i+tile_grid.coldiv)
            tile3_matches = tile.matchTileBorders(tile3, tile_grid)
        except:
            tile3_matches = {}
            
        axs[i-1].imshow(img)
        # axs[i-1].set_title(f"with {i+1}: {list(tile2_matches.keys())} ; with {i + tile_grid.coldiv}: {list(tile3_matches.keys())}")
        axs[i-1].set_xticks([])
        axs[i-1].set_yticks([])

    plt.show()

if __name__ == '__main__':
    grid = tileGrid(4,4,1100,800, pattern = "snake_by_row")
    print(grid)
    # grid.getNeighbouringTiles(2)

    # plotLabeledImageOverlap(glob("../out_dir/labeled1_MERFISH_nuclei_tile*.tif"), grid)

